(********************************************************************************)
(*  ocamlcore-api: Access to OCamlCore.org services                             *)
(*                                                                              *)
(*  Copyright (C) 2011, OCamlCore SARL                                          *)
(*                                                                              *)
(*  This library is free software; you can redistribute it and/or modify it     *)
(*  under the terms of the GNU Lesser General Public License as published by    *)
(*  the Free Software Foundation; either version 2.1 of the License, or (at     *)
(*  your option) any later version, with the OCaml static compilation           *)
(*  exception.                                                                  *)
(*                                                                              *)
(*  This library is distributed in the hope that it will be useful, but         *)
(*  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  *)
(*  or FITNESS FOR A PARTICULAR PURPOSE. See the file COPYING for more          *)
(*  details.                                                                    *)
(*                                                                              *)
(*  You should have received a copy of the GNU Lesser General Public License    *)
(*  along with this library; if not, write to the Free Software Foundation,     *)
(*  Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA               *)
(********************************************************************************)

(** Session management in connection with FusionForge 
    @author Sylvain Le Gall
  *)

open CalendarLib
open Lwt

type t = 
    {
      sess_hash:       string;
      sess_user_id:    int;
      sess_time:       Calendar.t;
      sess_ip:         Unix.inet_addr;
      sess_user_agent: string;
    }

let stub_data: (t OCAStubData.t) = 
  OCAStubData.create ()

let sep = "-*-"

let to_string t = 
  Printf.sprintf 
    "user_id: %d; time: %s; ip: %s; user_agent: %s; session_hash: %s;\n"
    t.sess_user_id 
    (Printer.Calendar.to_string t.sess_time)
    (Unix.string_of_inet_addr t.sess_ip)
    t.sess_user_agent t.sess_hash

let compare t1 t2 = 
  let mk_tuple t = 
    (t.sess_hash,
     t.sess_user_id,
     t.sess_ip,
     t.sess_user_agent)
  in
    match compare (mk_tuple t1) (mk_tuple t2) with
      | 0 ->
          Calendar.compare t1.sess_time t2.sess_time
      | n ->
          n

let cookie_name = "session_ser"

let to_cookie t =
  let session_serial = 
    String.concat sep 
      [string_of_int t.sess_user_id;
       Int64.to_string
         (Int64.of_float 
            (Calendar.to_unixfloat t.sess_time));
       Unix.string_of_inet_addr t.sess_ip;
       t.sess_user_agent]
  in
    Netencoding.Url.encode
      ((Netencoding.Base64.encode session_serial)^sep^t.sess_hash)

let of_cookie cookie = 
  let split_sep str = 
    BatString.nsplit str "-*-" 
  in
  let session_serial, session_hash = 
    match split_sep (Netencoding.Url.decode cookie) with 
      | [fst; snd] ->
          fst, snd
      | _ ->
          failwith "Malformed session string"
  in
  let user_id, time, ip, user_agent = 
    match split_sep (Netencoding.Base64.decode session_serial) with 
      | user_id :: time :: ip :: tl ->
          int_of_string user_id, 
          Calendar.from_unixfloat (float_of_string time),
          Unix.inet_addr_of_string ip, 
          String.concat sep tl
      | _ ->
          failwith "Malformed session serial string"
  in
    {
      sess_hash        = session_hash;
      sess_user_id     = user_id;
      sess_time        = time;
      sess_ip          = ip;
      sess_user_agent  = user_agent;
    }

let check ~ctxt ?user_agent ?ip t = 
  let user_agent_check = 
    match user_agent with 
      | Some str ->
          t.sess_user_agent = str
      | None ->
          true
  in
  let ip_check = 
    match ip with 
      | Some ip ->
          t.sess_ip = ip
      | None ->
          true
  in
    match user_agent_check && ip_check, ctxt with
      | true, `Live db ->
          begin
            OCASql.query db
              "SELECT count(*) FROM user_session WHERE \
                session_hash = $1 AND user_id = $2 AND ip_addr = $3"
              [Some t.sess_hash; 
               Some (OCASql.PGOCaml.string_of_int t.sess_user_id);
               Some (Unix.string_of_inet_addr t.sess_ip)]
            >>= function
              | [Some res] :: _ ->
                  return ((int_of_string res) > 0)
              | _->
                  return false
          end
      | true, `Stub stub_id ->
          catch 
            (fun () ->
               OCAStubData.get_one stub_data stub_id 
                 (fun t' -> compare t' t = 0)
               >>= fun _ ->
               return true)
            (function
               | Not_found ->
                   return false
               | e ->
                   fail e)

      | false, _ ->
          return false

